/* Copyright (C) 2009 Registro.br. All rights reserved. 
* 
* Redistribution and use in source and binary forms, with or without 
* modification, are permitted provided that the following conditions are 
* met:
* 1. Redistribution of source code must retain the above copyright 
*    notice, this list of conditions and the following disclaimer.
* 2. Redistributions in binary form must reproduce the above copyright
*    notice, this list of conditions and the following disclaimer in the
*    documentation and/or other materials provided with the distribution.
* 
* THIS SOFTWARE IS PROVIDED BY REGISTRO.BR ``AS IS'' AND ANY EXPRESS OR
* IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
* WARRANTIE OF FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO
* EVENT SHALL REGISTRO.BR BE LIABLE FOR ANY DIRECT, INDIRECT,
* INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
* BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
* OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
* ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
* TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
* USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
* DAMAGE.
 */
package br.registro.dnsshim.xfrd.dao.filesystem;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileWriter;
import java.io.IOException;
import java.util.HashMap;
import java.util.Map;
import java.util.Properties;
import java.util.Map.Entry;

import org.apache.log4j.Logger;

import br.registro.dnsshim.common.repository.FileSystemUtil;
import br.registro.dnsshim.xfrd.domain.TsigKeyInfo;
import br.registro.dnsshim.xfrd.repository.TsigKeyInfoRepository;

public class TsigKeyInfoDao implements TsigKeyInfoRepository {
	private static final Logger logger = Logger.getLogger(TsigKeyInfoDao.class);

	private static final String EXTENSION = ".properties";
	private static final String SEP_SECRET = "#";

	private static File file;

	public TsigKeyInfoDao() {
		StringBuilder root = new StringBuilder();
		try {
			// {DNSSHIM_HOME|user.home}/dnsshim/conf/xfrd/tsig-key.properties
			if (System.getenv("DNSSHIM_HOME") == null) {
				root.append(System.getProperty("user.home"));
				root.append(File.separatorChar);
				root.append("dnsshim");
			} else {
				root.append(System.getenv("DNSSHIM_HOME"));
			}
			root.append(File.separatorChar);
			root.append("xfrd");
			root.append(File.separatorChar);
			root.append("conf");
			root.append(File.separatorChar);

			File directory = new File(root.toString());
			FileSystemUtil.makePath(directory);

			root.append("tsig-key");
			root.append(EXTENSION);

			file = new File(root.toString());
			if (!file.exists()) {
				if (!file.createNewFile()) {
					throw new IOException("Cannot create file:"+file);
				}
			}

			if (logger.isDebugEnabled()) {
				logger.debug("XFRD tsig-key file:" + file);
			}

		} catch (IOException e) {
			logger.error(e.getMessage(), e);
		}
	}

	@Override
	public Map<String, TsigKeyInfo> loadAll() throws IOException {
		if (logger.isInfoEnabled()) {
			logger.info("Loading all TSIG keys");
		}

		Properties conf = new Properties();
		FileInputStream fis = new FileInputStream(file);
		conf.load(fis);

		Map<String, TsigKeyInfo> tsigKeys = new HashMap<String, TsigKeyInfo>();

		for (Object host : conf.keySet()) {
			TsigKeyInfo tsigKeyInfo = loadByIdentifier(conf, (String) host);
			if (tsigKeyInfo != null) {
				tsigKeys.put((String) host, tsigKeyInfo);
			}
		}
		fis.close();
		return tsigKeys;
	}

	@Override
	public TsigKeyInfo loadByIdentifier(String host) throws IOException {
		if (logger.isDebugEnabled()) {
			logger.debug("Loading TSIG keys for host " + host);
		}

		Properties conf = new Properties();
		FileInputStream fis = new FileInputStream(file);
		conf.load(fis);

		TsigKeyInfo tsigKeyInfo = loadByIdentifier(conf, host);
		fis.close();
		return tsigKeyInfo;
	}

	private TsigKeyInfo loadByIdentifier(Properties conf, String host) {
		String property = conf.getProperty(host);
		if (property == null) {
			logger.warn("TSIG key for host " + host + " not found");
			return null;
		}

		String[] parts = property.split(";");
		TsigKeyInfo tsigKeyInfo = new TsigKeyInfo();
		tsigKeyInfo.setHost(host);
		Map<String, String> keys = tsigKeyInfo.getKeys();
		for (String key : parts) {
			String[] keyParts = key.split(SEP_SECRET);
			if (keyParts.length != 2) {
				logger.error("Invalid format in TSIG key file");
				return null;
			}
			
			String keyName = keyParts[0];
			String secret = keyParts[1];
			keys.put(keyName, secret);
		}
		
		return tsigKeyInfo;
	}

	@Override
	public void save(TsigKeyInfo tsigKeyInfo) throws IOException {
		if (logger.isDebugEnabled()) {
			logger.debug("Saving TSIG key...");
		}

		Properties conf = new Properties();
		FileInputStream fis = new FileInputStream(file);
		conf.load(fis);

		Map<String, String> keys = tsigKeyInfo.getKeys();
		StringBuilder value = new StringBuilder();
		for (Entry<String, String> entry : keys.entrySet()) {
			value.append(entry.getKey()).append(SEP_SECRET);
			value.append(entry.getValue()).append(";");
		}
		
		if (value.toString().isEmpty()) {
			conf.remove(tsigKeyInfo.getHost());
		} else {
			conf.put(tsigKeyInfo.getHost(), value.toString());
		}
		
		FileWriter fw = new FileWriter(file);
		conf.store(fw, "Generated by DNSSHIM");
		
		fis.close();
		fw.close();
	}

}
