/* Copyright (C) 2009 Registro.br. All rights reserved. 
* 
* Redistribution and use in source and binary forms, with or without 
* modification, are permitted provided that the following conditions are 
* met:
* 1. Redistribution of source code must retain the above copyright 
*    notice, this list of conditions and the following disclaimer.
* 2. Redistributions in binary form must reproduce the above copyright
*    notice, this list of conditions and the following disclaimer in the
*    documentation and/or other materials provided with the distribution.
* 
* THIS SOFTWARE IS PROVIDED BY REGISTRO.BR ``AS IS'' AND ANY EXPRESS OR
* IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
* WARRANTIE OF FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO
* EVENT SHALL REGISTRO.BR BE LIABLE FOR ANY DIRECT, INDIRECT,
* INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
* BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
* OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
* ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
* TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
* USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
* DAMAGE.
 */
package br.registro.dnsshim.xfrd.dao.filesystem;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileWriter;
import java.io.IOException;
import java.text.ParseException;
import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.Date;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;
import java.util.Properties;

import org.apache.log4j.Logger;

import br.registro.dnsshim.common.repository.FileSystemUtil;
import br.registro.dnsshim.util.ByteUtil;
import br.registro.dnsshim.util.DomainNameUtil;
import br.registro.dnsshim.xfrd.domain.DnskeyStatus;
import br.registro.dnsshim.xfrd.domain.ZoneInfo;
import br.registro.dnsshim.xfrd.repository.ZoneInfoRepository;

public class ZoneInfoDao implements ZoneInfoRepository {
	
	private static final Logger logger = Logger.getLogger(ZoneInfoDao.class);
	private static final String EXTENSION = ".properties";
	private final String BASE_DIR;

	public ZoneInfoDao() {
		StringBuilder root = new StringBuilder();
		try {
			// {DNSSHIM_HOME|user.home}/dnsshim/xfrd/zones/SHA1#1/SHA1#2/SHA1#3/<zones>.properties
			if (System.getenv("DNSSHIM_HOME") == null) {
				root.append(System.getProperty("user.home"));
				root.append(File.separatorChar);
				root.append("dnsshim");
			} else {
				root.append(System.getenv("DNSSHIM_HOME"));
			}
			root.append(File.separatorChar);
			root.append("xfrd");
			root.append(File.separatorChar);
			root.append("zones");
			root.append(File.separatorChar);

			File directory = new File(root.toString());
			FileSystemUtil.makePath(directory);
			
			if (logger.isDebugEnabled()) {
				logger.debug("ZoneInfo base dir:"+root);
			}
			
		} catch (IOException e) {
			logger.error(e.getMessage(), e);
		}
		BASE_DIR = root.toString();
		
	}

	@Override
	public void save(ZoneInfo zoneInfo) throws IOException {
		String zn = DomainNameUtil.trimDot(zoneInfo.getZonename());
		StringBuilder path = new StringBuilder(BASE_DIR);
		path.append(FileSystemUtil.hash(zn));
		
		File directory = new File(path.toString());
		FileSystemUtil.makePath(directory);
		
		path.append(zn);
		path.append(EXTENSION);
		
		if (logger.isDebugEnabled()) {
			logger.debug("Saving " + zn + " configurations");
		}
		
		File zoneFile = new File(path.toString());
		
		if (!zoneFile.exists()) {
			if (!zoneFile.createNewFile()){
				logger.error("Cannot create file:"+zoneFile);
			}
		}

		Properties properties = new Properties();
		//salvo em long para ficar unsigned...
		properties.put("max_increments", String.valueOf(ByteUtil.toUnsigned(zoneInfo.getMaxIncrements())));
		properties.put("expiration", String.valueOf(ByteUtil.toUnsigned(zoneInfo.getExpiration())));
		properties.put("soa_minimum_ttl", String.valueOf(ByteUtil.toUnsigned(zoneInfo.getSoaMinimum())));
		properties.put("ttl_dnskey", String.valueOf(ByteUtil.toUnsigned(zoneInfo.getTtlDnskey())));
		properties.put("ttl_zone", String.valueOf(ByteUtil.toUnsigned(zoneInfo.getTtlZone())));
		properties.put("zonename", zoneInfo.getZonename());
		SimpleDateFormat dateFormat = new SimpleDateFormat("yyyyMMddHHmmss");
		Date nextPublication = new Date(zoneInfo.getNextPublication());
		properties.put("next_publication", dateFormat.format(nextPublication));

		StringBuilder keyNames = new StringBuilder();
		for (String keyIdentifier : zoneInfo.getKeys(DnskeyStatus.SIGN)) {
			keyNames.append(keyIdentifier);
			keyNames.append(';');
		}
		properties.put("keys_sign", keyNames.toString());

		keyNames = new StringBuilder();
		for (String keyIdentifier : zoneInfo.getKeys(DnskeyStatus.PUBLISH)) {
			keyNames.append(keyIdentifier);
			keyNames.append(';');
		}
		properties.put("keys_publish", keyNames.toString());

		keyNames = new StringBuilder();
		for (String keyIdentifier : zoneInfo.getKeys(DnskeyStatus.NONE)) {
			keyNames.append(keyIdentifier);
			keyNames.append(';');
		}
		properties.put("keys_none", keyNames.toString());

		keyNames = new StringBuilder();
		for (String keyName : zoneInfo.getOldKeys(DnskeyStatus.SIGN)) {
			keyNames.append(keyName);
			keyNames.append(';');
		}
		properties.put("last_keys_sign", keyNames.toString());

		keyNames = new StringBuilder();
		for (String keyName : zoneInfo.getOldKeys(DnskeyStatus.PUBLISH)) {
			keyNames.append(keyName);
			keyNames.append(';');
		}
		properties.put("last_keys_publish", keyNames.toString());

		keyNames = new StringBuilder();
		for (String keyName : zoneInfo.getOldKeys(DnskeyStatus.NONE)) {
			keyNames.append(keyName);
			keyNames.append(';');
		}
		properties.put("last_keys_none", keyNames.toString());
		
		properties.put("signed", String.valueOf(zoneInfo.isSigned()));
		
		StringBuilder slaves = new StringBuilder();
		for (String group : zoneInfo.getSlaveGroups()) {
			slaves.append(group);
			slaves.append(';');
		}
		properties.put("slaves", slaves.toString());
		
		StringBuilder users = new StringBuilder();
		for (String username : zoneInfo.getUsers()) {
			users.append(username);
			users.append(';');
		}
		properties.put("users", users.toString());

		FileWriter fw = new FileWriter(zoneFile);
		properties.store(fw, "Generated by DNSSHIM");
		fw.close();
		

		if (logger.isDebugEnabled()) {
			logger.debug(zn + " configurations saved @ "+zoneFile);
		}
	}
	
	public void delete(String zonename) throws IOException {
		StringBuilder path = new StringBuilder(BASE_DIR);
		path.append(FileSystemUtil.hash(zonename));
		File dir = new File(path.toString());
		path.append(zonename).append(EXTENSION);
		File file = new File(path.toString());
		file.delete();
		while (dir.list().length == 0) {
			File parent = dir.getParentFile();
			dir.delete();
			dir = parent;
		}
	}

	@Override
	public ZoneInfo findByZonename(String zonename) throws IOException {
		if (zonename.isEmpty()) {
			throw new IllegalArgumentException("Empty zone name");
		}
		
		String zn = DomainNameUtil.trimDot(zonename);
		StringBuilder path = new StringBuilder(BASE_DIR);
		path.append(FileSystemUtil.hash(zn));
		path.append(zn);
		path.append(EXTENSION);

		if (logger.isDebugEnabled()) {
			logger.debug("Finding " + zn + " configurations");
		}

		File zoneFile = new File(path.toString());
		if (zoneFile.exists()) {
			ZoneInfo zoneInfo = loadProperty(new File(path.toString()));
			return zoneInfo;
		}
		return null;
	}

	@Override
	public Map<String, ZoneInfo> findAll() throws IOException {
		if (logger.isInfoEnabled()) {
			logger.info("Loading all zone configurations...");
		}
		File rootDirectory = new File(BASE_DIR);
		Map<String, ZoneInfo> zones = new LinkedHashMap<String, ZoneInfo>();
		loadZonesRecursively(rootDirectory, zones);
		return zones;
	}

	private void loadZonesRecursively(File rootDirectory, Map<String, ZoneInfo> zones) throws IOException {
		File[] files = rootDirectory.listFiles();
		for (File file : files) {
			if (file.isDirectory()) {
				loadZonesRecursively(file, zones);
			} else {
				if (file.getName().endsWith(EXTENSION)) {
					ZoneInfo zoneInfo = loadProperty(file);
					if (logger.isDebugEnabled()) {
						logger.debug(zoneInfo.getZonename() + " configurations loaded");
					}
					zones.put(zoneInfo.getZonename(), zoneInfo);
				}
			}
		}
	}

	private ZoneInfo loadProperty(File file) throws IOException {
		if (logger.isDebugEnabled()) {
			logger.debug("Loading property file:"+file);
		}
		Properties conf = new Properties();
		FileInputStream fis = new FileInputStream(file);
		conf.load(fis);
		
		// saved as long to be unsigned
		int maxIncrements = (int)Long.parseLong(conf.getProperty("max_increments", String.valueOf(ZoneInfo.DEFAULT_MAX_INCREMENTS)));
		int expiration = (int)Long.parseLong(conf.getProperty("expiration", String.valueOf(ZoneInfo.DEFAULT_EXPIRATION)));
		int soa_minimum = (int)Long.parseLong(conf.getProperty("soa_minimum_ttl", String.valueOf(ZoneInfo.DEFAULT_SOA_MINIMUM)));
		int ttl_dnskey = (int)Long.parseLong(conf.getProperty("ttl_dnskey", String.valueOf(ZoneInfo.DEFAULT_TTL_DNSKEY)));
		int ttl_zone = (int)Long.parseLong(conf.getProperty("ttl_zone", String.valueOf(ZoneInfo.DEFAULT_TTL_ZONE)));
		String next_publication = conf.getProperty("next_publication", "19700101000000");
		boolean signed = Boolean.parseBoolean(conf.getProperty("signed", String.valueOf(ZoneInfo.DEFAULT_SIGNED)));
		
		
		String[] signKeys = conf.getProperty("keys_sign","").split(";");
		String[] publishKeys = conf.getProperty("keys_publish","").split(";");
		String[] noneKeys = conf.getProperty("keys_none","").split(";");
		String[] oldSignKeys = conf.getProperty("last_keys_sign","").split(";");
		String[] oldPublishKeys = conf.getProperty("last_keys_publish","").split(";");
		String[] oldNoneKeys = conf.getProperty("last_keys_none","").split(";");
		String[] slaves = conf.getProperty("slaves","").split(";");
		String[] users = conf.getProperty("users","").split(";");
		
		ZoneInfo zoneInfo = new ZoneInfo();

		zoneInfo.setMaxIncrements(maxIncrements);
		zoneInfo.setExpiration(expiration);
		zoneInfo.setSoaMinimum(soa_minimum);
		zoneInfo.setTtlDnskey(ttl_dnskey);
		zoneInfo.setTtlZone(ttl_zone);
		zoneInfo.setZonename(conf.getProperty("zonename", ZoneInfo.DEFAULT_ZONENAME));
		zoneInfo.setSigned(signed);
		
		SimpleDateFormat dateFormat = new SimpleDateFormat("yyyyMMddHHmmss");
		Date nextPub = null;
		try {
			nextPub = dateFormat.parse(next_publication);
		} catch (ParseException e) {
			nextPub = new Date();
			nextPub.setTime(0);
		}
		zoneInfo.setNextPublication(nextPub.getTime());
		
		for (String keyIdentifier : signKeys) {
			String id = keyIdentifier.trim();
			if (!id.isEmpty()) {
				zoneInfo.addKey(id, DnskeyStatus.SIGN);
			}
		}

		for (String keyIdentifier : publishKeys) {
			String id = keyIdentifier.trim();
			if (!id.isEmpty()) {
				zoneInfo.addKey(id, DnskeyStatus.PUBLISH);
			}
		}

		for (String keyIdentifier : noneKeys) {
			String id = keyIdentifier.trim();
			if (!id.isEmpty()) {
				zoneInfo.addKey(id, DnskeyStatus.NONE);
			}
		}

		for (String keyIdentifier : oldSignKeys) {
			String keyName = keyIdentifier.trim();
			if (!keyName.isEmpty()) {
				zoneInfo.addOldKey(keyName, DnskeyStatus.SIGN);
			}
		}

		for (String keyIdentifier : oldPublishKeys) {
			String keyName = keyIdentifier.trim();
			if (!keyName.isEmpty()) {
				zoneInfo.addOldKey(keyName, DnskeyStatus.PUBLISH);
			}
		}

		for (String keyIdentifier : oldNoneKeys) {
			String keyName = keyIdentifier.trim();
			if (!keyName.isEmpty()) {
				zoneInfo.addOldKey(keyName, DnskeyStatus.NONE);
			}
		}

		for (String slave : slaves) {
			String s = slave.trim();
			if (!s.isEmpty()) {
				zoneInfo.assignSlaveGroup(s);
			}
		}
		
		for (String username : users) {
			String u = username.trim();
			if (!u.isEmpty()) {
				zoneInfo.addUser(u);
			}
		}
		
		fis.close();
		
		// always keep a well-formed config file
		save(zoneInfo);
		return zoneInfo;
	}

	@Override
	public List<String> findAllZonenames() throws IOException {
		List<String> zonenames = new ArrayList<String>();
		File rootDirectory = new File(BASE_DIR);
		loadZonenames(rootDirectory, zonenames);
		return zonenames;
	}
	
	private void loadZonenames(File rootDirectory, List<String> zonenames) {
		File[] files = rootDirectory.listFiles();
		for (File file : files) {
			if (file.isDirectory()) {
				loadZonenames(file, zonenames);
			} else {
				if (file.getName().endsWith(EXTENSION)) {
					String identifier =
						file.getName().substring(0, file.getName().lastIndexOf('.'));
					zonenames.add(identifier);
				}
			}
		}
	}
}