/* Copyright (C) 2009 Registro.br. All rights reserved. 
* 
* Redistribution and use in source and binary forms, with or without 
* modification, are permitted provided that the following conditions are 
* met:
* 1. Redistribution of source code must retain the above copyright 
*    notice, this list of conditions and the following disclaimer.
* 2. Redistributions in binary form must reproduce the above copyright
*    notice, this list of conditions and the following disclaimer in the
*    documentation and/or other materials provided with the distribution.
* 
* THIS SOFTWARE IS PROVIDED BY REGISTRO.BR ``AS IS'' AND ANY EXPRESS OR
* IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
* WARRANTIE OF FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO
* EVENT SHALL REGISTRO.BR BE LIABLE FOR ANY DIRECT, INDIRECT,
* INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
* BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
* OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
* ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
* TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
* USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
* DAMAGE.
 */
package br.registro.dnsshim.xfrd.scheduler;

import java.io.IOException;
import java.util.Date;
import java.util.concurrent.PriorityBlockingQueue;

import org.apache.log4j.Logger;

import br.registro.dnsshim.xfrd.dao.filesystem.ZonenameDao;
import br.registro.dnsshim.xfrd.domain.ZoneInfo;
import br.registro.dnsshim.xfrd.domain.Zonename;
import br.registro.dnsshim.xfrd.domain.logic.ZoneInfoCache;
import br.registro.dnsshim.xfrd.repository.ZonenameRepository;
import br.registro.dnsshim.xfrd.scheduler.domain.ZoneResign;

public class ResignManager {
	private static final Logger logger = Logger.getLogger(ResignManager.class);

	private static final PriorityBlockingQueue<ZoneResign> queue = new PriorityBlockingQueue<ZoneResign>();
	
	static {
		Thread resignScheduler = new Thread(new ResignScheduler(queue));
		resignScheduler.start();
	}

	private ResignManager() {}
	
	public static void add(ZoneResign signature) {
		synchronized (queue) {
			ZoneInfoCache zoneInfoCache = ZoneInfoCache.getInstance();
			ZoneInfo zoneInfo = zoneInfoCache.get(signature.getZonename());
			zoneInfo.setNextPublication(signature.getExpiration());
			zoneInfoCache.put(zoneInfo);
			queue.put(signature);
			if (logger.isDebugEnabled()) {
				logger.debug("New resign scheduled for zone "+signature.getZonename()+" on "+signature.getExpiration());
				logger.debug("Notifying scheduler");
			}
			queue.notify();
		}
	}
	
	public boolean remove(ZoneResign signature) {
		synchronized (queue) {
			boolean r = queue.remove(signature);
			queue.notify();
			return r;
			
		}	
	}
	
	public static void clear() {
		synchronized (queue) {
			queue.clear();
			queue.notify();
		}	
	}
	
	public static boolean remove(String zonename) {
		synchronized (queue) {
			ZoneResign zoneResign = new ZoneResign();
			zoneResign.setZonename(zonename);
			return queue.remove(zoneResign);
		}	
	}
	
	public static void launch() {
		Thread resignLoader = new Thread(new Runnable() {
			public void run() {
				try {
					if (logger.isInfoEnabled()) {
						logger.info("DNSSHIM has been started.Verifying next expiration for all zones");
					}
					
					ZonenameRepository repository = new ZonenameDao();
					ZoneInfoCache zoneInfoCache = ZoneInfoCache.getInstance();
					for (Zonename zonename : repository.findAll()) {
						ZoneInfo zoneInfo = zoneInfoCache.get(zonename.getName());
						if (zoneInfo != null && zoneInfo.isSigned()) {
							// Tell scheduler to manage
							ZoneResign zoneResign = new ZoneResign();
							zoneResign.setZonename(zonename.getName());
							zoneResign.setExpiration(zoneInfo.getNextPublication());
							synchronized (queue) {
								queue.put(zoneResign);
								queue.notify();
							}
							if (logger.isDebugEnabled()) {
								logger.debug("New resign scheduled for zone "+zoneResign.getZonename()+" on "+ new Date(zoneResign.getExpiration()));
								logger.debug("Notifying scheduler");
							}
						}
					}
				} catch (IOException e) {
					logger.error(e.getMessage(), e);
				}
			}
		});
		resignLoader.start();
	}
}
